# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Dict, Tuple

from pandas import DataFrame, MultiIndex
import json
import os
from time import sleep
import warnings
import logging

from lomond import WebSocket
from ibm_watson_machine_learning.utils.autoai.enums import (
    PredictionType, RegressionAlgorithms, ClassificationAlgorithms, RunStateTypes, DataConnectionTypes)
from ibm_watson_machine_learning.utils.autoai.errors import PipelineNotLoaded, FitNeeded, InvalidPredictionType, \
    AutoAIComputeError, ForecastingUnsupportedOperation, NoAvailableMetrics, LibraryNotCompatible, \
    CannotConnectToWebsocket
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning.utils.autoai.progress_bar import ProgressBar
from ibm_watson_machine_learning.utils.autoai.utils import (
    fetch_pipelines, is_ipython, ProgressGenerator, check_dependencies_versions, try_import_autoai_libs,
    create_summary, prepare_auto_ai_model_to_publish_normal_scenario, get_sw_spec_and_type_based_on_sklearn)
from .base_engine import BaseEngine
from ibm_watson_machine_learning.helpers.connections import DataConnection

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace
    from ibm_watson_machine_learning.helpers.connections import DataConnection
    from sklearn.pipeline import Pipeline

__all__ = [
    "WMLEngine"
]

logging.getLogger('lomond').setLevel(logging.CRITICAL)


class WMLEngine(BaseEngine):
    """
    WML Engine provides unified API to work with AutoAI Pipelines trainings on WML.

    Parameters
    ----------
    workspace: WorkSpace, required
        WorkSpace object with wml client and all needed parameters.
    """

    def __init__(self, workspace: 'WorkSpace') -> None:
        self.workspace = workspace
        self._wml_client = workspace.wml_client
        self._auto_pipelines_parameters = None
        self._wml_pipeline_metadata = None
        self._wml_training_metadata = None
        self._wml_stored_pipeline_details = None
        self._current_run_id = None
        self._20_class_limit_removal_test = False

    def _initialize_wml_pipeline_metadata(self, **kwargs) -> None:
        """
        Initialization of WML Pipeline Document (WML client Meta Parameter) with provided parameters and default ones.
        """

        self._wml_pipeline_metadata = {
            self._wml_client.pipelines.ConfigurationMetaNames.NAME:
                self._auto_pipelines_parameters.get('name', 'Default name.'),
            self._wml_client.pipelines.ConfigurationMetaNames.DESCRIPTION:
                self._auto_pipelines_parameters.get('desc', 'Default description'),
            self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT: {
                'doc_type': 'pipeline',
                'version': '2.0',
                'pipelines': [{
                    'id': 'autoai',
                    'runtime_ref': 'hybrid',
                    'nodes': [{
                        'id': 'automl',
                        'type': 'execution_node',
                        'parameters': {
                            'stage_flag': True,
                            'output_logs': True,
                            'input_file_separator': self._auto_pipelines_parameters.get('csv_separator'),
                            'encoding': self._auto_pipelines_parameters.get('encoding'),
                            'drop_duplicates': self._auto_pipelines_parameters.get('drop_duplicates'),
                            'optimization': {
                                'learning_type': 'timeseries' if self._auto_pipelines_parameters.get('prediction_type') == PredictionType.FORECASTING else self._auto_pipelines_parameters.get('prediction_type'),
                                'max_num_daub_ensembles': self._auto_pipelines_parameters.get('max_num_daub_ensembles'),
                                'cognito_transform_names':
                                    self._auto_pipelines_parameters.get('cognito_transform_names'),
                                'run_cognito_flag': True,
                                'holdout_param': self._auto_pipelines_parameters.get('holdout_size')
                            }
                        },
                        'runtime_ref': 'autoai',
                        'op': 'kube'
                    }]
                }],
                'runtimes': [{
                    'id': 'autoai',
                    'name': 'auto_ai.kb' if self._auto_pipelines_parameters.get('prediction_type') != PredictionType.FORECASTING else 'auto_ai.ts',
                    'app_data': {
                        'wml_data': {
                            'hardware_spec': {}
                        }
                    }
                }],
                'primary_pipeline': 'autoai'
            }
        }

        if kwargs.get('enable_all_data_sources') is not None:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['enable_all_data_sources'
            ] = kwargs['enable_all_data_sources']

        # Text transformer params
        if not self._wml_client.ICP_30 and not self._wml_client.ICP_35:
            if self._auto_pipelines_parameters.get('text_processing') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'text_processing_flag'] = self._auto_pipelines_parameters.get('text_processing')

            if self._auto_pipelines_parameters.get('word2vec_feature_number') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'text_processing_options'] = {'word2vec':
                                                {'output_dim': self._auto_pipelines_parameters.get('word2vec_feature_number')}}
            if self._auto_pipelines_parameters.get('text_columns_names'):
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'text_columns'] = self._auto_pipelines_parameters.get('text_columns_names')

        # end of section

        # pass fairness info if not CPD
        if not self._wml_client.ICP:
            if self._auto_pipelines_parameters.get('fairness_info') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'fairness_info'] = self._auto_pipelines_parameters.get('fairness_info')
        # end of section

        # note: Additional parameters for benchmark subsampling with data detector
        if self._wml_client.ICP and self._wml_client.ICP_40:
            if self._auto_pipelines_parameters.get('sampling_type') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'sampling_type'] = self._auto_pipelines_parameters.get('sampling_type')

            if self._auto_pipelines_parameters.get('n_parallel_data_connections') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'n_parallel_data_connections'] = self._auto_pipelines_parameters.get('n_parallel_data_connections')

            if self._auto_pipelines_parameters.get('calculate_data_metrics') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'calculate_data_metrics'] = self._auto_pipelines_parameters.get('calculate_data_metrics')

            if self._auto_pipelines_parameters.get('logical_batch_size') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'logical_batch_size'] = self._auto_pipelines_parameters.get('logical_batch_size')

            if self._auto_pipelines_parameters.get('metrics_on_logical_batch') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'metrics_on_logical_batch'] = self._auto_pipelines_parameters.get('metrics_on_logical_batch')

            if self._auto_pipelines_parameters.get('batch_size') is not None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters'][
                    'batch_size'] = self._auto_pipelines_parameters.get('batch_size')
        # --- end note

        if self._auto_pipelines_parameters.get('prediction_type') == PredictionType.FORECASTING:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'target_columns'] = self._auto_pipelines_parameters['prediction_columns']
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'timestamp_column'] = self._auto_pipelines_parameters['timestamp_column_name']
            if self._auto_pipelines_parameters['backtest_num']:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'num_backtest'] = self._auto_pipelines_parameters['backtest_num']
            if self._auto_pipelines_parameters['lookback_window']:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'lookback_window'] = self._auto_pipelines_parameters['lookback_window']
            if self._auto_pipelines_parameters['forecast_window']:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'prediction_horizon'] = self._auto_pipelines_parameters['forecast_window']
            if self._auto_pipelines_parameters['backtest_gap_length'] != None:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'gap_len'] = self._auto_pipelines_parameters['backtest_gap_length']
        else:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'label'] = self._auto_pipelines_parameters['prediction_column']
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'compute_pipeline_notebooks_flag'] = self._auto_pipelines_parameters.get('notebooks', False)
            if 'scoring' in self._auto_pipelines_parameters:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'scorer_for_ranking'] = self._auto_pipelines_parameters.get('scoring')

        if self._20_class_limit_removal_test:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['pipelines'][0]['nodes'][0]['parameters'].update({
                'one_vs_all_file': True
            })

        # note: in new v4 api we have missing hw_spec name
        t_size = self._auto_pipelines_parameters['t_shirt_size']
        if 0 < len(t_size) <= 2:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                'app_data']['wml_data']['hardware_spec']['name'] = t_size.upper()

        else:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0][
                'app_data']['wml_data']['hardware_spec']['id'] = t_size
        # --- end note

        if self._auto_pipelines_parameters.get('train_sample_rows_test_size') is not None:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'train_sample_rows_test_size'] = self._auto_pipelines_parameters['train_sample_rows_test_size']

        if self._auto_pipelines_parameters.get('t_shirt_size') == 's':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 6e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'm':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 9e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'l':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 15e9

        elif self._auto_pipelines_parameters.get('t_shirt_size') == 'xl':
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_adaptive_subsampling_max_mem_usage'] = 25e9

        if self._auto_pipelines_parameters.get('include_only_estimators') is not None:
            # note: transform values from Enum
            try:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                    'daub_include_only_estimators'] = [
                    algorithm.value for algorithm in self._auto_pipelines_parameters.get('include_only_estimators')
                ]
            # note: if user pass strings instead of enums
            except AttributeError:
                pass

        # note: only pass positive label when scoring is binary
        if (self._auto_pipelines_parameters.get('positive_label')
                and self._auto_pipelines_parameters.get('prediction_type') == PredictionType.BINARY):
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'positive_label'] = self._auto_pipelines_parameters.get('positive_label')
        # --- end note

        # note: only pass daub_give_priority_to_runtime when is not None
        if self._auto_pipelines_parameters.get('daub_give_priority_to_runtime'):
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters']['optimization'][
                'daub_runtime_ranking_power'] = self._auto_pipelines_parameters.get('daub_give_priority_to_runtime')
        # --- end note

        # note: only pass excel_sheet when it is different than 0
        if self._auto_pipelines_parameters.get('excel_sheet') != 0:
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'][0]['parameters'][
                'excel_sheet'] = self._auto_pipelines_parameters.get('excel_sheet')
        # --- end note

        # note: KB part: send version autoai_pod_version if set in pipeline details details (See supported formats)
        if not self._wml_client.ICP and self._auto_pipelines_parameters['autoai_pod_version'] is not None:
            self._wml_pipeline_metadata[
                self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT]['runtimes'][0]['version'] = (
                self._auto_pipelines_parameters['autoai_pod_version'])
        # --- end note
        
        # note: OBM part, only if user specified OBM as a preprocess method
        if self._auto_pipelines_parameters['data_join_graph']:
            # note: remove kb params for OBM only training
            if self._auto_pipelines_parameters.get('data_join_only'):
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'] = []
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'runtimes'] = []
            # --- end note
            else:
                # note: update autoai KB POD metadata
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['inputs'] = [{"id": "kb_input",
                                                              "links": [{"node_id_ref": "obm",
                                                                         "port_id_ref": "obm_out"
                                                                         }]
                                                              }]

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['outputs'] = [{"id": "outputData"
                                                               }]

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['id'] = 'kb'

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['runtime_ref'] = 'kb'

                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'runtimes'][0]['id'] = 'kb'
                # --- end note

            # note: insert OBM experiment metadata into the first place of nodes
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'pipelines'][0]['nodes'].insert(0, self._auto_pipelines_parameters['data_join_graph'].to_dict())
            # --- end note

            # note: insert OBM runtime metadata into the first place
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'runtimes'].insert(0, {"id": "obm",
                                       "name": "auto_ai.obm",
                                       "app_data": {
                                           "wml_data": {
                                               "hardware_spec": {
                                                   "name": "M-Spark",
                                                   "num_nodes": self._auto_pipelines_parameters['data_join_graph'].worker_nodes_number
                                               }
                                           }
                                       }
                                       }
                                   )

            # note: in new v4 api we have missing hw_spec name
            t_size = self._auto_pipelines_parameters['data_join_graph'].t_shirt_size
            self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                'runtimes'][0]['app_data']['wml_data']['hardware_spec']['name'] = f"{t_size.upper()}-Spark"
            # --- end note
            # --- end note

            if not self._wml_client.ICP:
                self._wml_pipeline_metadata[
                    self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'runtimes'][0]['version'] = self._auto_pipelines_parameters['obm_pod_version']

            else:
                self._wml_pipeline_metadata[self._wml_client.pipelines.ConfigurationMetaNames.DOCUMENT][
                    'pipelines'][0]['nodes'][0]['parameters']['engine'][
                    'template_id'] = 'spark-3.0.0-automl-cp4d-template'
        # --- end note

    def _initialize_wml_training_metadata(self,
                                          training_data_reference: List['DataConnection'],
                                          training_results_reference: 'DataConnection') -> None:
        """
        Initialization of wml training metadata (WML client Meta Parameter).

        Parameters
        ----------
        training_data_reference: List[DataConnection], required
            Data storage connection details to inform where training data is stored.

        training_results_reference: DataConnection, required
            Data storage connection details to store pipeline training results.

        """
        self._wml_training_metadata = {
            self._wml_client.training.ConfigurationMetaNames.TAGS:
                ['autoai'],

            self._wml_client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES:
                [connection._to_dict() for connection in training_data_reference],

            self._wml_client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE:
                training_results_reference._to_dict(),

            self._wml_client.training.ConfigurationMetaNames.PIPELINE:
                {'id': self._wml_stored_pipeline_details['metadata']['id']},

            self._wml_client.training.ConfigurationMetaNames.DESCRIPTION:
                f"{self._auto_pipelines_parameters.get('name', 'Default name.')} - "
                f"wml pipeline: {self._wml_stored_pipeline_details['metadata']['id']}",

            self._wml_client.training.ConfigurationMetaNames.NAME:
                f"{self._auto_pipelines_parameters.get('desc', 'Default description')} - "
                f"wml pipeline: {self._wml_stored_pipeline_details['metadata']['id']}",
        }

        # note: Delete unnecessary keys for connection_asset in training_data_reference
        for connection in \
                self._wml_training_metadata[self._wml_client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES] +\
                [self._wml_training_metadata[self._wml_client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE]]:
            if connection.get("type") == DataConnectionTypes.CA:
                try:
                    del connection["connection"]["secret_access_key"]
                    del connection["connection"]["access_key_id"]
                    del connection["connection"]["endpoint_url"]
                except KeyError:
                    pass
        # --- end note

    ##########################################################
    #   WML Pipeline Part / AutoPipelineOptimizer Init Part  #
    ##########################################################
    def initiate_remote_resources(self, params: dict, **kwargs) -> None:
        """
        Initializes the AutoPipelines with supplied parameters.

        Parameters
        ----------
        params: dict, required
            AutoAi experiment parameters.
        """
        self._auto_pipelines_parameters = params
        self._initialize_wml_pipeline_metadata(**kwargs)
        if 'enable_all_data_sources' in kwargs:
            del kwargs['enable_all_data_sources']
        self._wml_stored_pipeline_details = self._wml_client.pipelines.store(meta_props=self._wml_pipeline_metadata, **kwargs)

    def get_params(self) -> dict:
        """
        Get configuration parameters of AutoPipelineOptimizer.

        Returns
        -------
        Dictionary with AutoPipelineOptimizer parameters.
        """
        if self._auto_pipelines_parameters is not None:
            self._auto_pipelines_parameters['run_id'] = self._current_run_id

        return self._auto_pipelines_parameters

    ##########################################################
    #   WML Training Part / AutoPipelineOptimizer Run Part   #
    ##########################################################
    def fit(self,
            training_data_reference: List['DataConnection'],
            training_results_reference: 'DataConnection',
            background_mode: bool = True) -> dict:
        """
        Run a training process on WML of autoai on top of the training data referenced by training_data_connection.

        Parameters
        ----------
        training_data_reference: List[DataConnection], required
            Data storage connection details to inform where training data is stored.

        training_results_reference: DataConnection, required
            Data storage connection details to store pipeline training results.

        background_mode: bool, optional
            Indicator if fit() method will run in background (async) or (sync).

        Returns
        -------
        Dictionary with run details.
        """
        self._initialize_wml_training_metadata(training_data_reference=training_data_reference,
                                               training_results_reference=training_results_reference)

        run_params = self._wml_client.training.run(meta_props=self._wml_training_metadata,
                                                   asynchronous=True)

        self._current_run_id = run_params['metadata'].get('id', run_params['metadata'].get('guid'))

        if background_mode:
            return self._wml_client.training.get_details(training_uid=self._current_run_id)

        else:
            wml_pipeline_details = self.get_params()
            number_of_estimators = wml_pipeline_details.get('max_num_daub_ensembles', 2)

            base_url = self._wml_client.service_instance._href_definitions.get_trainings_href()
            url = f"{base_url.replace('https', 'wss')}/{self._current_run_id}"

            params = self._wml_client._params()

            if self._wml_client.default_project_id:
                url = f"{url}?project_id={self._wml_client.default_project_id}&version={params.get('version')}"
            elif self._wml_client.default_space_id:
                url = f"{url}?space_id={self._wml_client.default_space_id}&version={params.get('version')}"

            gen = ProgressGenerator()
            end = False

            progress_bar = ProgressBar(desc="Total", total=gen.get_total(), position=0, ncols=100, bar_format='{l_bar}{bar}| [{elapsed}<{remaining}, ' '{rate_fmt}{postfix}]')

            progress_bar.set_description(desc='Started waiting for resources')

            attempts_no = 10
            disconnection_no = 0

            while not end:
                websocket = WebSocket(url)
                websocket.add_header(bytes("Authorization", "utf-8"),
                                     bytes("Bearer " + self._wml_client.service_instance._get_token(), "utf-8"))

                try:
                    for event in websocket.connect():
                        try:
                            websocket.send_text('Ping')
                        except:
                            pass
                        if event.name == 'text':
                            status = json.loads(event.text)['entity']['status']

                            try:
                                process = status['metrics'][0]['context']['intermediate_model']['process']
                            except:
                                process = ''

                            if status.get('state') == RunStateTypes.FAILED:
                                sleep(3)

                                message = status.get('message', {}).get('text', 'Training failed')
                                progress_bar.set_description(desc=message)

                                if 'Not Found' in str(message) or 'HeadObject' in str(message):
                                    raise AutoAIComputeError(self._current_run_id,
                                                             reason=f"Fetching training data error. Please check if COS credentials, "
                                                                    f"bucket name and path are correct or file system path is correct.")

                                elif 'Failed_Image' in str(message):
                                    raise AutoAIComputeError(
                                        self._current_run_id,
                                        reason=f"This version of WML V4 SDK with AutoAI component is no longer supported. "
                                               f"Please update your package to the newest version. "
                                               f"Error: {message}")

                                else:
                                    raise AutoAIComputeError(self._current_run_id, reason=f"Error: {message}")
                            elif 'message' in status and 'text' in status['message']:
                                text = status['message']['text']

                                if self._current_run_id in text and 'completed' in text:
                                    end = True
                                    progress_bar.set_description(desc=text)
                                    progress_bar.last_update()
                                    progress_bar.close()
                                    break
                                elif 'AutoAI model computed' not in text:
                                    prepared_text = text.split(': ')[-1].split("global:")[-1]
                                    max_len = 40

                                    if len(prepared_text) > max_len:
                                        progress_bar.set_description(desc=prepared_text[:(max_len - 3)] + '...')
                                    else:
                                        progress_bar.set_description(
                                            desc=prepared_text + ' ' * (max_len - len(prepared_text)))

                                progress_bar.increment_counter(gen.get_progress(process))
                                progress_bar.update()
                            elif process:
                                progress_bar.increment_counter(gen.get_progress(process))
                                progress_bar.update()
                        elif event.name in ['closing', 'closed']:
                            end = True
                            websocket.close()
                            break
                        elif event.name in ['disconnected']:
                            disconnection_no += 1
                            websocket.close()
                            break
                except:
                    websocket.close()
                    break

                if disconnection_no >= attempts_no:
                    raise CannotConnectToWebsocket(attempts_no)

            return self._wml_client.training.get_details(training_uid=self._current_run_id)

    def get_run_status(self) -> str:
        """
        Check status/state of initialized AutoPipelineOptimizer run if ran in background mode

        Returns
        -------
        Dictionary with run status details.
        """

        if self._current_run_id is None:
            raise FitNeeded(reason="Firstly schedule a fit job by using the fit() method.")

        return self._wml_client.training.get_status(training_uid=self._current_run_id).get('state')

    def get_run_details(self) -> dict:
        """
        Get fit/run details.

        Returns
        -------
        Dictionary with AutoPipelineOptimizer fit/run details.
        """

        if self._current_run_id is None:
            raise FitNeeded(reason="Firstly schedule a fit job by using the fit() method.")

        details = self._wml_client.training.get_details(training_uid=self._current_run_id)

        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']
            return details
        else:
            return details

    def cancel_run(self) -> None:
        """Cancels an AutoAI run."""
        if self._current_run_id is None:
            raise FitNeeded(reason="To cancel an AutoAI run, first schedule a fit job by using the fit() method.")

        self._wml_client.training.cancel(training_uid=self._current_run_id)

    #################################################################################
    #   WML Auto_AI Trained Pipelines Part / AutoPipelineOptimizer Pipelines Part   #
    #################################################################################
    def summary(self) -> 'DataFrame':
        """
        Prints AutoPipelineOptimizer Pipelines details (autoai trained pipelines).

        Returns
        -------
        Pandas DataFrame with computed pipelines and ML metrics.
        """
        if self._current_run_id is None:
            raise FitNeeded(reason="To list computed pipelines, first schedule a fit job by using the fit() method.")

        details = self._wml_client.training.get_details(training_uid=self._current_run_id)
        return create_summary(details=details, scoring=self._auto_pipelines_parameters['scoring'])

    def get_pipeline_details(self, pipeline_name: str = None) -> dict:
        """
        Fetch specific pipeline details, eg. steps etc.

        Parameters
        ----------
        pipeline_name: str, optional
            Pipeline name eg. Pipeline_1, if not specified, best pipeline parameters will be fetched

        Returns
        -------
        Dictionary with pipeline parameters.
        """

        if self._current_run_id is None:
            raise FitNeeded(reason="To list computed pipelines parameters, "
                                   "first schedule a fit job by using a fit() method.")

        if pipeline_name is None:
            pipeline_name = self.summary().index[0]
        run_params = self._wml_client.training.get_details(training_uid=self._current_run_id)

        pipeline_parameters = {
            "composition_steps": [],
            "pipeline_nodes": [],
        }

        # note: retrieve all additional pipeline evaluation information
        for pipeline in run_params['entity']['status'].get('metrics', []):
            if pipeline['context']['intermediate_model']['name'].split('P')[-1] == pipeline_name.split('_')[-1]:

                pipeline_parameters['composition_steps'] = pipeline['context']['intermediate_model'][
                    'composition_steps']
                pipeline_parameters['pipeline_nodes'] = pipeline['context']['intermediate_model']['pipeline_nodes']
                if 'timeseries' not in pipeline['context']:
                    pipeline_parameters['ml_metrics'] = self._get_metrics(pipeline)
                else:
                    if not 'ts_metrics' in pipeline_parameters:
                        pipeline_parameters['ts_metrics'] = {}
                    pipeline_parameters['ts_metrics'].update(self._get_metrics(pipeline))

                if (self._auto_pipelines_parameters['prediction_type'] == 'binary'
                        or (self._auto_pipelines_parameters['prediction_type'] == 'classification'
                            and 'binary_classification' in str(run_params))):

                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)
                    pipeline_parameters['confusion_matrix'] = self._get_confusion_matrix(pipeline, run_params)
                    pipeline_parameters['roc_curve'] = self._get_roc_curve(pipeline, run_params)

                elif (self._auto_pipelines_parameters['prediction_type'] == 'multiclass'
                      or (self._auto_pipelines_parameters['prediction_type'] == 'classification'
                          and 'multi_class_classification' in str(run_params))):

                    one_vs_all_data = self._get_data_from_property_or_file('one_vs_all',
                                                   pipeline['context'].get('multi_class_classification', {'one_vs_all': []}),
                                                   run_params['entity']['results_reference'], [{}, {}])

                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)
                    pipeline_parameters['roc_curve'] = self._get_one_vs_all(one_vs_all_data, run_params, 'roc_curve')
                    pipeline_parameters['confusion_matrix'] = self._get_one_vs_all(one_vs_all_data, run_params, 'confusion_matrix')

                elif self._auto_pipelines_parameters['prediction_type'] == 'regression':
                    pipeline_parameters['features_importance'] = self._get_features_importance(pipeline)

                elif self._auto_pipelines_parameters['prediction_type'] == 'forecasting':
                    data = self._get_data_from_property_or_file('daub_status', pipeline['context']['timeseries'],
                                                                     run_params['entity']['results_reference'],
                                                                     [{}, {}])
                    if 'visualization' in data:
                        if 'visualization' not in pipeline_parameters:
                            pipeline_parameters['visualization'] = {}
                        pipeline_parameters['visualization'].update(data['visualization'])

                else:
                    raise InvalidPredictionType(self._auto_pipelines_parameters['prediction_type'],
                                                reason="Prediction type not recognized.")

                break

        # --- end note

        return pipeline_parameters

    @staticmethod
    def _get_metrics(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve evaluation metrics data from particular pipeline details."""
        is_ml_metrics = 'ml_metrics' in pipeline_details
        is_ts_metrics = 'ts_metrics' in pipeline_details

        if not is_ml_metrics and not is_ts_metrics:
            raise NoAvailableMetrics()

        if is_ml_metrics:
            data = pipeline_details['ml_metrics']
        elif is_ts_metrics:
            data = pipeline_details['ts_metrics']

        columns = [field for field in data.keys()]
        values = [[value for value in data.values()]]

        metrics = DataFrame(data=values, columns=columns)
        metrics.index = ['score']
        metrics = metrics.transpose()

        return metrics

    def _get_data_from_property_or_file(self, property_name: str, details: Dict, results_reference: Dict, default_value: List) -> List:
        if f'{property_name}_location' in details:
            path = details[f'{property_name}_location']
            conn = DataConnection._from_dict(results_reference)
            conn._wml_client = self._wml_client
            return conn._download_json_file(path)
        elif property_name in details:  # TODO probably should be removed in future
            return details[property_name]
        else:
            return default_value

    def _get_confusion_matrix(self, pipeline_details: Dict, run_params: Dict) -> 'DataFrame':
        """Retrieve confusion matrix data from particular pipeline details or from file. (Binary)"""
        data = self._get_data_from_property_or_file('confusion_matrix',
                                                         pipeline_details['context'].get('binary_classification', {}),
                                                         run_params['entity']['results_reference'],
                                                         [{}, {}])
        columns = [name for name in data[0].keys()]
        values = [[value for value in item.values()] for item in data]

        confusion_matrix_data = DataFrame(data=values, columns=columns)
        confusion_matrix_data.index = confusion_matrix_data['true_class']
        confusion_matrix_data.drop(['true_class'], axis=1, inplace=True)

        return confusion_matrix_data

    @staticmethod
    def _get_features_importance(pipeline_details: Dict) -> 'DataFrame':
        """Retrieve features importance data from particular pipeline details."""
        data = pipeline_details['context'].get('features_importance', [{'features': {}}])[0]['features']
        columns = [name for name in data.keys()]
        values = [[value for value in data.values()]]

        features_importance_data = DataFrame(data=values, columns=columns, index=['features_importance'])
        features_importance_data = features_importance_data.transpose()
        features_importance_data.sort_values(by='features_importance', ascending=False, inplace=True)

        return features_importance_data

    def _get_roc_curve(self, pipeline_details: Dict, run_params: Dict) -> 'DataFrame':
        """Retrieve roc curve data from particular pipeline details or from file. (Binary)"""
        data = self._get_data_from_property_or_file('roc_curve',
                                                         pipeline_details['context'].get('binary_classification', {}),
                                                         run_params['entity']['results_reference'],
                                                         [])

        true_classes = [item['true_class'] for item in data]
        values = [data[0]['fpr'], data[1]['fpr'], data[0]['thresholds'], data[1]['thresholds'], data[0]['tpr'],
                  data[1]['tpr']]

        index = MultiIndex.from_product([['fpr', 'thresholds', 'tpr'], true_classes],
                                        names=['measurement types', 'true_class'])

        roc_curve = DataFrame(data=values, index=index)

        return roc_curve

    def _get_one_vs_all(self, one_vs_all_data: List, run_params: Dict, option: str) -> 'DataFrame':
        """Retrieve roc curve or confusion matrix data from particular pipeline details. (Multiclass)"""

        one_vs_all_data = [x for x in one_vs_all_data if 'multi-class' not in x]

        if option == 'confusion_matrix':
            confusion_matrix_data = [self._get_data_from_property_or_file('confusion_matrix', item,
                                     run_params['entity']['results_reference'], [{}, {}]) for item in one_vs_all_data]

            columns = ['fn', 'fp', 'tn', 'tp', 'true_class']
            confusion_matrix_values = [[item[c] for c in columns] for item in confusion_matrix_data]
            confusion_matrix_data = DataFrame(data=confusion_matrix_values,
                                              columns=columns)
            confusion_matrix_data.index = confusion_matrix_data['true_class']
            confusion_matrix_data.drop(['true_class'], axis=1, inplace=True)

            return confusion_matrix_data

        elif option == 'roc_curve':
            roc_curve_data = [self._get_data_from_property_or_file('roc_curve', item,
                              run_params['entity']['results_reference'], []) for item in one_vs_all_data]

            true_classes = [item['true_class'] for item in roc_curve_data]
            values = [item['fpr'] for item in roc_curve_data] + [item['thresholds'] for item in
                                                                    roc_curve_data] + [item['tpr'] for item in roc_curve_data]
            index = MultiIndex.from_product([['fpr', 'thresholds', 'tpr'], true_classes],
                                            names=['measurement types', 'true_class'])
            roc_curve = DataFrame(data=values, index=index)

            return roc_curve

        else:
            raise ValueError(f"Wrong option for \"one_vs_all\": {option}. "
                             f"Available options: [confusion_matrix, roc_curve]")

    def get_pipeline(self, pipeline_name: str, local_path: str = '.',
                     persist: 'bool' = False) -> Tuple['Pipeline', bool]:
        """
        Download specified pipeline from WML.

        Parameters
        ----------
        pipeline_name: str, required
            Pipeline name, if you want to see the pipelines names, please use summary() method.

        local_path: str, optional
            Local filesystem path, if not specified, current directory is used.

        persist: bool, optional
            Indicates if selected pipeline should be stored locally.

        Returns
        -------
        Scikit-Learn pipeline.
        """
        try_import_autoai_libs()

        # note: Show warning if OBM
        if 'auto_ai.obm' in str(self._wml_stored_pipeline_details):
            warnings.warn("OBM pipeline can be used only for inspection and deployment purposes.", Warning, stacklevel=2)
        # end note

        run_params = self._wml_client.training.get_details(training_uid=self._current_run_id)

        # note: recreation of s3 creds from connection asset
        results_reference = DataConnection._from_dict(run_params['entity']['results_reference'])
        results_reference._wml_client = self._wml_client
        results_reference._check_if_connection_asset_is_s3()
        run_params['entity']['results_reference'] = results_reference._to_dict()
        # --- end note

        if 'timeseries' in run_params['entity']['status']['metrics'][0]['context']:
            raise ForecastingUnsupportedOperation()

        try:
            pipelines, check_lale = fetch_pipelines(
                run_params=run_params,
                path=local_path,
                pipeline_name=pipeline_name,
                load_pipelines=True,
                store=persist,
                wml_client=self._wml_client
            )

        except LibraryNotCompatible as e:
            raise e

        except Exception as e:
            raise PipelineNotLoaded(
                pipeline_name,
                reason=f'Fit finished but there was some error during loading a pipeline from WML. Error: {e}')

        return pipelines.get(pipeline_name), check_lale

    def get_best_pipeline(self, local_path: str = '.', persist: 'bool' = False) -> Tuple['Pipeline', bool]:
        """
        Download best pipeline from WML.

        Parameters
        ----------
        local_path: str, optional
            Local filesystem path, if not specified, current directory is used.

        persist: bool, optional
            Indicates if selected pipeline should be stored locally.

        Returns
        -------
        Scikit-Learn pipeline.
        """
        best_pipeline_name = self.summary().index[0]
        return self.get_pipeline(best_pipeline_name)



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import requests,json, time
from ibm_watson_machine_learning.utils import SPACES_IMPORTS_DETAILS_TYPE, SPACES_EXPORTS_DETAILS_TYPE, SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos, print_text_header_h2
from ibm_watson_machine_learning.utils import StatusLogger, print_text_header_h1, print_text_header_h2
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import  WMLClientError
from ibm_watson_machine_learning.metanames import SpacesPlatformMetaNames, SpacesPlatformMemberMetaNames
from ibm_watson_machine_learning.instance_new_plan import ServiceInstanceNewPlan


_DEFAULT_LIST_LENGTH = 50

class PlatformSpaces(WMLResource):
    """
    Store and manage your spaces
    """
    ConfigurationMetaNames = SpacesPlatformMetaNames()
    MemberMetaNames = SpacesPlatformMemberMetaNames()
    """MetaNames for spaces creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._client = client

    def _get_resources(self, url, op_name, params=None):
        if params is not None and 'limit' in params.keys():
            if params[u'limit'] < 1:
                raise WMLClientError('Limit cannot be lower than 1.')
            elif params[u'limit'] > 1000:
                raise WMLClientError('Limit cannot be larger than 1000.')

        if len(params) > 0:
            if not self._ICP:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    params=params
                )
            else:
                response_get = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    params=params,
                    verify=False
                )

            return self._handle_response(200, op_name, response_get)
        else:

            resources = []

            while True:
                if not self._ICP:
                    response_get = requests.get( url, headers=self._client._get_headers())
                else:
                    response_get = requests.get( url, headers=self._client._get_headers(), verify=False)

                result = self._handle_response(200, op_name, response_get)
                resources.extend(result['resources'])

                if 'next' not in result:
                    break
                else:
                    url = self._wml_credentials["url"]+result['next']['href']
                    if('start=invalid' in url):
                        break
            return {
                "resources": resources
            }

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props, background_mode=True):
        """
                Create a space. The instance associated with the space via COMPUTE will be used for billing purposes on
                cloud. Note that STORAGE and COMPUTE are applicable only for cloud

                **Parameters**

                .. important::
                    #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                    >>> client.spaces.ConfigurationMetaNames.get()

                      **type**: dict\n

                    #. **background_mode**:  Indicator if store() method will run in background (async) or (sync). Default: True
                      **type**: bool\n

                **Output**

                .. important::

                    **returns**: metadata of the stored space\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>>  client.spaces.ConfigurationMetaNames.NAME: 'my_space',
                 >>>  client.spaces.ConfigurationMetaNames.DESCRIPTION: 'spaces',
                 >>>  client.spaces.ConfigurationMetaNames.STORAGE: {"resource_crn": "provide crn of the COS storage"},
                 >>>  client.spaces.ConfigurationMetaNames.COMPUTE: {"name": "test_instance",
                 >>>                                                 "crn": "provide crn of the instance"}
                 >>> }
                 >>> spaces_details = client.spaces.store(meta_props=metadata)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        PlatformSpaces._validate_type(meta_props, u'meta_props', dict, True)

        if ('compute' in meta_props or 'storage' in meta_props) and self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' and 'COMPUTE' meta props are not applicable on "
                                 "IBM Cloud Pak® for Data. If using any of these, remove and retry")

        if 'storage' not in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' is mandatory for cloud")

        if 'compute' in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            if 'name' not in meta_props[u'compute'] or 'crn' not in meta_props[u'compute']:
                raise WMLClientError("'name' and 'crn' is mandatory for 'COMPUTE'")
            temp_meta = meta_props[u'compute']
            temp_meta.update({'type': 'machine_learning'})

            meta_props[u'compute'] = temp_meta

        space_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        if 'compute' in meta_props and self._client.CLOUD_PLATFORM_SPACES:
            payload_compute = []
            payload_compute.append(space_meta[u'compute'])
            space_meta[u'compute'] = payload_compute

        if not self._ICP:
            creation_response = requests.post(
                self._client.service_instance._href_definitions.get_platform_spaces_href(),
                headers=self._client._get_headers(),
                json=space_meta)
        else:
            creation_response = requests.post(
                self._client.service_instance._href_definitions.get_platform_spaces_href(),
                headers=self._client._get_headers(),
                json=space_meta,
                verify=False)

        spaces_details = self._handle_response(202, u'creating new spaces', creation_response)

        # Cloud Convergence: Set self._client.wml_credentials['instance_id'] to instance_id
        # during client.set.default_space since that's where space is associated with client
        # and also in client.set.default_project
        #
        if 'compute' in spaces_details['entity'].keys() and self._client.CLOUD_PLATFORM_SPACES:
            instance_id = spaces_details['entity']['compute'][0]['guid']
            self._client.wml_credentials[u'instance_id'] = instance_id
            self._client.service_instance = ServiceInstanceNewPlan(self._client)
            self._client.service_instance.details = self._client.service_instance.get_details()


        if background_mode:
            print("Space has been created. However some background setup activities might still be on-going. "
                  "Check for 'status' field in the response. It has to show 'active' before space can be used. "
                  "If its not 'active', you can monitor the state with a call to spaces.get_details(space_id)")
            return spaces_details

        else:
            # note: monitor space status
            space_id = self.get_id(spaces_details)
            print_text_header_h1(u'Synchronous space creation with id: \'{}\' started'.format(space_id))

            status = spaces_details['entity']['status'].get('state')

            with StatusLogger(status) as status_logger:
                while status not in ['failed', 'error', 'completed', 'canceled', 'active']:
                    time.sleep(10)
                    spaces_details = self.get_details(space_id)
                    status = spaces_details['entity']['status'].get('state')
                    status_logger.log_state(status)
            # --- end note

            if u'active' in status:
                print_text_header_h2(u'\nCreating space  \'{}\' finished successfully.'.format(space_id))
            else:
                raise WMLClientError(
                    f"Space {space_id} creation failed with status: {spaces_details['entity']['status']}")

            return spaces_details

    @staticmethod
    def get_id(space_details):
        """
            Get space_id from space details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored space\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: space ID\n
                **return type**: str

            **Example**

             >>> space_details = client.spaces.store(meta_props)
             >>> space_id = client.spaces.get_id(space_details)
        """

        PlatformSpaces._validate_type(space_details, u'space_details', object, True)

        return WMLResource._get_required_element_from_dict(space_details, u'space_details',
                                                           [u'metadata', u'id'])


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(space_details):
        """
                Get Unique Id of the space. This method is deprecated. Use 'get_id(space_details)' instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the space\n
                      **type**: dict\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of space\n
                    **return type**: str\n

                **Example**

                 >>> space_details = client.spaces.store(meta_props)
                 >>> space_uid = client.spaces.get_uid(space_details)

        """
        PlatformSpaces._validate_type(space_details, u'space_details', object, True)

        return WMLResource._get_required_element_from_dict(space_details, u'space_details',
                                                           [u'metadata', u'id'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, space_id):
        """
            Delete a stored space.

            **Parameters**

            .. important::
                #. **space_uid**:  space ID\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.spaces.delete(deployment_id)
        """

        space_id = str_type_conv(space_id)
        PlatformSpaces._validate_type(space_id, u'space_id', STR_TYPE, True)

        space_endpoint = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        if not self._ICP:
            response_delete = requests.delete(space_endpoint, headers=self._client._get_headers())
        else:
            response_delete = requests.delete(space_endpoint, headers=self._client._get_headers(), verify=False)

        response = self._handle_response(202, u'space deletion', response_delete, False)

        print('DELETED')

        return response

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, space_id=None, limit=None, asynchronous=False, get_all=False):
        """
           Get metadata of stored space(s)

           **Parameters**

           .. important::
                #. **space_id**: Space ID \n
                   **type**: str\n
                #. **limit**: Applicable when space_id is not provided. If space_id is provided, this will be ignored \n
                   **type**: str\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                **returns**: metadata of stored space(s)\n
                **return type**: dict

           **Example**

            >>> space_details = client.spaces.get_details(space_uid)
            >>> space_details = client.spaces.get_details(limit=100)
            >>> space_details = client.spaces.get_details(limit=100, get_all=True)
            >>> space_details = []
            >>> for entry in client.spaces.get_details(limit=100, asynchronous=True, get_all=True):
            >>>    space_details.extend(entry)
        """

        space_id = str_type_conv(space_id)
        # PlatformSpaces._validate_type(space_id, u'space_id', STR_TYPE, True)
        PlatformSpaces._validate_type(space_id, u'space_id', STR_TYPE, False)

        href = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        if space_id is not None:
            if not self._ICP:
                response_get = requests.get(href, headers=self._client._get_headers())
            else:
                response_get = requests.get(href, headers=self._client._get_headers(), verify=False)

            return self._handle_response(200, 'Get space', response_get)

        else:
            return self._get_with_or_without_limit(self._client.service_instance._href_definitions.get_platform_spaces_href(),
                                                   limit,
                                                   'spaces',
                                                   summary=False,
                                                   pre_defined=False,
                                                   skip_space_project_chk=True,
                                                   _async=asynchronous,
                                                   _all=get_all)

    def list(self, limit=None, member=None, roles=None):
        """
           List stored spaces. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n
                #. **member**:  Filters the result list to only include spaces where the user with a matching user id
                                is a member\n
                   **type**: string\n
                #. **roles**:  limit number of fetched records\n
                   **type**: string\n

           **Output**

           .. important::
                This method only prints the list of all spaces in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.spaces.list()
        """

        PlatformSpaces._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_platform_spaces_href()

        params = {}

        if limit is not None:
            params.update({'limit': limit})

        if limit is None:
            params.update({'limit': 50})

        if member is not None:
            params.update({'member': member})

        if roles is not None:
            params.update({'roles': roles})

        space_resources = self._get_resources(href, 'spaces', params)[u'resources']

        # space_resources = self._get_no_space_artifact_details(href, None, limit, 'spaces')[u'resources']

        space_values = [(m[u'metadata'][u'id'],
                         m[u'entity'][u'name'],
                         m[u'metadata'][u'created_at']) for m in space_resources]

        if limit is None:
            print("Note: 'limit' is not provided. Only first 50 records will be displayed if the number of records "
                  "exceed 50")

        self._list(space_values, [u'ID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update(self, space_id, changes):
        """
                Updates existing space metadata. 'STORAGE' cannot be updated
                STORAGE and COMPUTE are applicable only for cloud

                **Parameters**

                .. important::
                    #. **space_uid**:  ID of space which definition should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::
                    **returns**: metadata of updated space\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.spaces.ConfigurationMetaNames.NAME:"updated_space",
                 >>> client.spaces.ConfigurationMetaNames.COMPUTE: {"name": "test_instance",
                 >>>                                                "crn": "v1:staging:public:pm-20-dev:us-south:a/09796a1b4cddfcc9f7fe17824a68a0f8:f1026e4b-77cf-4703-843d-c9984eac7272::"
                 >>>                                               }
                 >>> }
                 >>> space_details = client.spaces.update(space_id, changes=metadata)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        if ('compute' in changes or 'storage' in changes) and self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("'STORAGE' and 'COMPUTE' meta props are not applicable on"
                                 "IBM Cloud Pak® for Data. If using any of these, remove and retry")

        if 'storage' in changes:
            raise WMLClientError("STORAGE cannot be updated")

        space_id = str_type_conv(space_id)
        self._validate_type(space_id, u'space_id', STR_TYPE, True)
        self._validate_type(changes, u'changes', dict, True)
        meta_props_str_conv(changes)

        details = self.get_details(space_id)

        if 'compute' in changes and self._client.CLOUD_PLATFORM_SPACES:
            changes[u'compute'][u'type'] = 'machine_learning'

            payload_compute = []
            payload_compute.append(changes[u'compute'])
            changes[u'compute'] = payload_compute

        print("changes in update: ", changes)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes)

        print("patch payload: ", patch_payload)

        href = self._client.service_instance._href_definitions.get_platform_space_href(space_id)

        if not self._ICP:
            response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())
        else:
            response = requests.patch(href, json=patch_payload, headers=self._client._get_headers(), verify=False)

        updated_details = self._handle_response(200, u'spaces patch', response)

        # Cloud Convergence
        if 'compute' in updated_details['entity'].keys() and self._client.CLOUD_PLATFORM_SPACES:
            instance_id = updated_details['entity']['compute'][0]['guid']
            self._client.wml_credentials[u'instance_id'] = instance_id
            self._client.service_instance = ServiceInstanceNewPlan(self._client)
            self._client.service_instance.details = self._client.service_instance.get_details()

        return updated_details


#######SUPPORT FOR SPACE MEMBERS

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_member(self, space_id, meta_props):
        """
                Create a member within a space.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the member configuration. To see available meta names use:\n
                                    >>> client.spaces.MemberMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the stored member\n
                    **return type**: dict\n

                .. note::
                    * 'role' can be any one of the following "viewer, editor, admin"\n
                    * 'type' can be any one of the following "user,service"\n
                    * 'id' can be either service-ID or IAM-userID\n

                **Example**

                 >>> metadata = {
                 >>>  client.spaces.MemberMetaNames.MEMBERS: [{"id":"IBMid-100000DK0B", "type": "user", "role": "admin" }]
                 >>> }
                 >>> members_details = client.spaces.create_member(space_id=space_id, meta_props=metadata)

                 >>> metadata = {
                 >>>  client.spaces.MemberMetaNames.MEMBERS: [{"id":"iam-ServiceId-5a216e59-6592-43b9-8669-625d341aca71", "type": "service", "role": "admin" }]
                 >>> }
                 >>> members_details = client.spaces.create_member(space_id=space_id, meta_props=metadata)
        """

        space_id = str_type_conv(space_id)
        self._validate_type(space_id, u'space_id', STR_TYPE, True)

        PlatformSpaces._validate_type(meta_props, u'meta_props', dict, True)

        meta = {}

        if 'members' in meta_props:
            meta = meta_props
        elif 'member' in meta_props:
            dictionary = meta_props['member']
            payload = []
            payload.append(dictionary)
            meta['members'] = payload

        space_meta = self.MemberMetaNames._generate_resource_metadata(
            meta,
            with_validation=True,
            client=self._client
        )

        if not self._ICP:
            creation_response = requests.post(
                self._client.service_instance._href_definitions.get_platform_spaces_members_href(space_id),
                headers=self._client._get_headers(),
                json=space_meta)
        else:
            creation_response = requests.post(
                self._client.service_instance._href_definitions.get_platform_spaces_members_href(space_id),
                headers=self._client._get_headers(),
                json=space_meta,
                verify=False)

        # TODO: Change response code one they change it to 201
        members_details = self._handle_response(200, u'creating new members', creation_response)

        return members_details

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_member_details(self, space_id, member_id):
        """
           Get metadata of member associated with a space

           **Parameters**

           .. important::
                #. **space_id**: member ID \n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: metadata of member of a space\n
                **return type**: dict

           **Example**

            >>> member_details = client.spaces.get_member_details(space_uid,member_id)
        """

        space_id = str_type_conv(space_id)
        PlatformSpaces._validate_type(space_id, u'space_id', STR_TYPE, True)

        member_id = str_type_conv(member_id)
        PlatformSpaces._validate_type(member_id, u'member_id', STR_TYPE, True)

        href = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id, member_id)

        if not self._ICP:
            response_get = requests.get(href, headers=self._client._get_headers())
        else:
            response_get = requests.get(href, headers=self._client._get_headers(), verify=False)

        return self._handle_response(200, 'Get space member', response_get)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete_member(self, space_id, member_id):
        """
            Delete a member associated with a space.

            **Parameters**

            .. important::
                #. **space_id**:  space UID\n
                   **type**: str\n
                #. **member_id**:  member UID\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.spaces.delete_member(space_id,member_id)
        """

        space_id = str_type_conv(space_id)
        PlatformSpaces._validate_type(space_id, u'space_id', STR_TYPE, True)

        member_id = str_type_conv(member_id)
        PlatformSpaces._validate_type(member_id, u'member_id', STR_TYPE, True)

        member_endpoint = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id, member_id)

        if not self._ICP:
            response_delete = requests.delete(member_endpoint, headers=self._client._get_headers())
        else:
            response_delete = requests.delete(member_endpoint, headers=self._client._get_headers(), verify=False)

        print('DELETED')

        return self._handle_response(204, u'space member deletion', response_delete, False)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_member(self, space_id, member_id, changes):
            """
                    Updates existing member metadata.

                    **Parameters**

                    .. important::
                        #. **space_id**:  ID of space\n
                           **type**: str\n
                        #. **member_id**:  ID of member that needs to be updated\n
                           **type**: str\n
                        #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                           **type**: dict\n

                    **Output**

                    .. important::
                        **returns**: metadata of updated member\n
                        **return type**: dict\n

                    **Example**

                     >>> metadata = {
                     >>>  client.spaces.MemberMetaNames.MEMBER: {"role": "editor"}
                     >>> }
                     >>> member_details = client.spaces.update_member(space_id, member_id, changes=metadata)
            """

            space_id = str_type_conv(space_id)
            self._validate_type(space_id, u'space_id', STR_TYPE, True)
            member_id = str_type_conv(member_id)
            self._validate_type(member_id, u'member_id', STR_TYPE, True)

            self._validate_type(changes, u'changes', dict, True)
            meta_props_str_conv(changes)

            details = self.get_member_details(space_id, member_id)

            # The member record is a bit different than most other type of records we deal w.r.t patch
            # There is no encapsulating object for the fields. We need to be consistent with the way we
            # provide the meta in create/patch. When we give with .MEMBER, _generate_patch_payload
            # will generate with /member patch. So, separate logic for member patch inline here
            changes1 = changes['member']

            # Union of two dictionaries. The one in changes1 will override existent ones in current meta
            details.update(changes1)

            id_str = {}
            role_str = {}
            type_str = {}
            state_str = {}

            # if 'id' in details:
            #     id_str["op"] = "replace"
            #     id_str["path"] = "/id"
            #     id_str["value"] = details[u'id']
            if 'role' in details:
                role_str["op"] = "replace"
                role_str["path"] = "/role"
                role_str["value"] = details[u'role']
            # if 'type' in details:
            #     type_str["op"] = "replace"
            #     type_str["path"] = "/type"
            #     type_str["value"] = details[u'type']
            if 'state' in details:
                state_str["op"] = "replace"
                state_str["path"] = "/state"
                state_str["value"] = details[u'state']

            patch_payload = []

            # if id_str:
            #     patch_payload.append(id_str)
            if role_str:
                patch_payload.append(role_str)
            # if type_str:
            #     patch_payload.append(type_str)
            if state_str:
                patch_payload.append(state_str)

             # patch_payload = self.MemberMetaNames._generate_patch_payload(details, changes, with_validation=True)

            href = self._client.service_instance._href_definitions.get_platform_spaces_member_href(space_id,member_id)

            if not self._ICP:
                response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())
            else:
                response = requests.patch(href, json=patch_payload, headers=self._client._get_headers(), verify=False)

            updated_details = self._handle_response(200, u'members patch', response)

            return updated_details

    def list_members(self, space_id, limit=None, identity_type=None, role=None, state=None):
            """
               List stored members of a space. If limit is set to None there will be only first 50 records shown.

               **Parameters**

               .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n
                #. **identity_type**:  Find the member by type\n
                   **type**: string\n
                #. **role**:  Find the member by role\n
                   **type**: string\n
                #. **state**:  Find the member by state\n
                   **type**: string\n

               **Output**

               .. important::
                    This method only prints the list of all members associated with a space in a table format.\n
                    **return type**: None\n

               **Example**

                >>> client.spaces.list_members(space_id)
            """

            space_id = str_type_conv(space_id)
            self._validate_type(space_id, u'space_id', STR_TYPE, True)

            params = {}

            if limit is not None:
                params.update({'limit': limit})

            if limit is None:
                params.update({'limit': 50})

            if identity_type is not None:
                params.update({'type': identity_type})

            if role is not None:
                params.update({'role': role})

            if state is not None:
                params.update({'state': state})

            href = self._client.service_instance._href_definitions.get_platform_spaces_members_href(space_id)

            member_resources = self._get_resources(href, 'space members', params)[u'resources']

            # space_values = [(m[u'metadata'][u'id'],
            #                  m[u'entity'][u'id'],
            #                  m[u'entity'][u'type'],
            #                  m[u'entity'][u'role'],
            #                  m[u'entity'][u'state'],
            #                  m[u'metadata'][u'created_at']) for m in member_resources]


            # self._list(space_values, [u'ID', u'IDENTITY',
            #                           u'IDENTITY_TYPE',
            #                           u'ROLE',
            #                           u'STATE',
            #                           u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

            space_values = [(m[u'id'],
                             m[u'type'],
                             m[u'role'],
                             m[u'state']) if 'state' in m else
                            (m[u'id'],
                             m[u'type'],
                             m[u'role'],
                             None) for m in member_resources]

            if limit is None:
                print("Note: 'limit' is not provided. Only first 50 records will be displayed if the number of records "
                      "exceed 50")

            self._list(space_values, [u'ID',
                                      u'TYPE',
                                      u'ROLE',
                                      u'STATE'], limit, _DEFAULT_LIST_LENGTH)








# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import requests

from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import ConnectionMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os
import json

_DEFAULT_LIST_LENGTH = 0


class Connections(WMLResource):
    """
    Store and manage your Connections.
    """
    ConfigurationMetaNames = ConnectionMetaNames()
    """MetaNames for Connection creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'connection_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {'id': response_data['metadata']['asset_id'],
                                       'space_id': response_data['metadata']['space_id'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'create_time': response_data['metadata']['create_time'],
                                       'last_access_time': response_data['metadata']['usage'].get('last_access_time')
                                       },
                          'entity': {
                              'datasource_type': response_data['entity']['datasource_type'],
                              'description': response_data['entity']['description'],
                              'name': response_data['entity']['name'],
                              'origin_country': response_data['entity']['origin_country'],
                              'owner_id': response_data['entity']['owner_id'],
                              'properties': response_data['entity']['properties']
                                }
                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:
                    # TODO :  There is no owner id in connection api response of WSD1.0, this might need a check in WSD2.0
                    # kaganesa: add the below line in 2.0
                    #  'owner_id': response_data['entity']['owner_id'],
                    new_el = {'metadata': {'id': response_data['metadata']['asset_id'],
                                           'project_id': response_data['metadata']['project_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'create_time': response_data['metadata']['create_time'],
                                           'last_access_time': response_data['metadata']['usage'].get('last_access_time')
                                           },
                              'entity': {
                                  'datasource_type': response_data['entity']['datasource_type'],
                                  'description': response_data['entity']['description'],
                                  'name': response_data['entity']['name'],
                                  'origin_country': response_data['entity']['origin_country'],
                                 # 'owner_id': response_data['entity']['owner_id'],
                                  'properties': response_data['entity']['properties']
                               }
                              }


                else:
                    new_el = {'metadata': {'id': response_data['metadata']['asset_id'],
                                           'project_id': response_data['metadata']['project_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'create_time': response_data['metadata']['create_time'],
                                           'last_access_time': response_data['metadata']['usage'].get('last_access_time')
                                           },
                              'entity': {
                                  'datasource_type': response_data['entity']['datasource_type'],
                                  'description': response_data['entity']['description'],
                                  'name': response_data['entity']['name'],
                                  'origin_country': response_data['entity']['origin_country'],
                                  'owner_id': response_data['entity']['owner_id'],
                                  'properties': response_data['entity']['properties']
                                  }
                              }
                    if 'href' in response_data['metadata']:
                        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                            href_without_host = response_data['href'].split('.com')[-1]
                            new_el[u'metadata'].update({'href': href_without_host})
                        else:
                            new_el['metadata'].update({'href': response_data['href']})

            return new_el
        except Exception as e:
            raise WMLClientError(Messages.get_message(response_data, message_id="failed_to_read_response_from_down_stream_service"))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, connection_id=None):
        """
        Get connection details for the given unique Connection id. If no connection_id is passed, details for all connections will be returned.

        **Parameters**

        .. important::
            #. **connection_id**: Unique id of Connection\n
                **type**: str\n

        **Output**

        .. important::
            **returns**: Metadata of the stored Connection\n
            **return type**: dict\n

        **Example**

         >>> connection_details = client.connections.get_details(connection_id)
         >>> connection_details = client.connections.get_details()
        """
        self._client._check_if_either_is_set()
        Connections._validate_type(connection_id, u'connection_id', STR_TYPE, False)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        verify = not self._client.ICP_30 and not self._client.ICP and not self._client.WSD
        if connection_id:
            response = requests.get(self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id),
                                    params=self._client._params(),
                                    headers=header_param, verify=verify)

            if response.status_code == 200:
                return self._get_required_element_from_response(
                    self._handle_response(200, u'get connection details', response))
            else:
                return self._handle_response(200, u'get connection details', response)
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_connections_href(),
                                    params=self._client._params(),
                                    headers=header_param, verify=verify)

            if response.status_code == 200:
                return {'resources': [self._get_required_element_from_response(x) for x in response.json()['resources']]}
            else:
                return self._handle_response(200, u'get connection details', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create(self, meta_props):
        """
                Creates a connection. Input to PROPERTIES field examples

                1. MySQL
                      >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                      >>>   "database": "database",
                      >>>   "password": "password",
                      >>>   "port": "3306",
                      >>>   "host": "host url",
                      >>>   "ssl": "false",
                      >>>   "username": "username"
                      >>> }

                2. Google Big query

                   a. Method1: Use service account json. The service account json generated can be provided as
                               input as-is. Provide actual values in json. Example is only indicative to show
                               the fields. Refer to Google big query documents how to generate the service account json\n

                      >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                      >>>     "type": "service_account",
                      >>>     "project_id": "project_id",
                      >>>     "private_key_id": "private_key_id",
                      >>>     "private_key": "private key contents",
                      >>>     "client_email": "client_email",
                      >>>     "client_id": "client_id",
                      >>>     "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                      >>>     "token_uri": "https://oauth2.googleapis.com/token",
                      >>>     "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                      >>>     "client_x509_cert_url": "client_x509_cert_url"
                      >>> }

                   b. Method2: Using OAuth Method. Refer to Google big query documents how to generate OAuth token\n

                       >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                       >>>    "access_token": "access token generated for big query",
                       >>>    "refresh_token": "refresh token",
                       >>>    "project_id": "project_id",
                       >>>    "client_secret": "This is your gmail account password",
                       >>>   "client_id": "client_id"
                       >>> }

                3. MS SQL
                     >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                     >>>      "database": "database",
                     >>>      "password": "password",
                     >>>      "port": "1433",
                     >>>      "host": "host",
                     >>>      "username": "username"
                     >>> }

                4. Tera data
                     >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                     >>>      "database": "database",
                     >>>      "password": "password",
                     >>>      "port": "1433",
                     >>>      "host": "host",
                     >>>      "username": "username"
                     >>> }


                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the connection configuration. To see available meta names use:\n
                            >>> client.connections.ConfigurationMetaNames.get()

                      **type**: dict\n
                **Output**

                .. important::

                    **returns**: metadata of the stored connection\n
                    **return type**: dict\n

                **Example**

                 >>> sqlserver_data_source_type_id = client.connections.get_datasource_type_uid_by_name('sqlserver')
                 >>> connections_details = client.connections.create({
                 >>> client.connections.ConfigurationMetaNames.NAME: "sqlserver connection",
                 >>> client.connections.ConfigurationMetaNames.DESCRIPTION: "connection description",
                 >>> client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: sqlserver_data_source_type_id,
                 >>> client.connections.ConfigurationMetaNames.PROPERTIES: { "database": "database",
                 >>>                                                         "password": "password",
                 >>>                                                         "port": "1433",
                 >>>                                                         "host": "host",
                 >>>                                                         "username": "username"}
                 >>> })

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        connection_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        big_query_data_source_type_id = self.get_datasource_type_uid_by_name('bigquery')

        # Either service acct json credentials can be given or oauth json can be given
        # If service acct json, then we need to create a newline json with "credentials" key
        if connection_meta[u'datasource_type'] == big_query_data_source_type_id:
            if 'private_key' in connection_meta[u'properties']:
                result = json.dumps(connection_meta[u'properties'], separators=(',\n', ':'))
                newmap = {"credentials": result}
                connection_meta[u'properties'] = newmap

        connection_meta.update({'origin_country': 'US'})
        #Step1  : Create an asset
        print(Messages.get_message(message_id="creating_connections"))

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if not self._client.ICP_30 and not self._client.ICP and not self._client.WSD:
            creation_response = requests.post(
                    self._client.service_instance._href_definitions.get_connections_href(),
                    headers=header_param,
                    params = self._client._params(),
                    json=connection_meta
            )
        else:
            creation_response = requests.post(
                self._client.service_instance._href_definitions.get_connections_href(),
                headers=header_param,
                json=connection_meta,
                params=self._client._params(),
                verify=False
            )
        connection_details = self._handle_response(201, u'creating new connection', creation_response)
        if creation_response.status_code == 201:
            connection_id = connection_details["metadata"]["asset_id"]
            print(Messages.get_message(message_id="success"))
            return self._get_required_element_from_response(connection_details)
        else:
            raise WMLClientError(Messages.get_message(message_id="failed_while_creating_connections"))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, connection_id):
        """
            Delete a stored Connection.

            **Parameters**

            .. important::
                #. **connection_id**: Unique id of the connection to be deleted.\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.connections.delete(connection_id)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        pipeline_uid = str_type_conv(connection_id)
        Connections._validate_type(connection_id, u'connection_id', STR_TYPE, True)
        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        connection_endpoint = self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id)
        if not self._ICP and not self._client.ICP_30 and not self._client.WSD:
            response_delete = requests.delete(connection_endpoint, params=self._client._params(),
                                              headers=header_param)
        else:
            response_delete = requests.delete(connection_endpoint, params=self._client._params(),
                                              headers=header_param, verify=False)

        return self._handle_response(204, u'connection deletion', response_delete, False)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(connection_details):
        """
                Get Unique Id of stored connection.

                **Parameters**

                .. important::

                   #. **connection_details**:  Metadata of the stored connection\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored connection\n
                    **return type**: str\n

                **Example**


                 >>> connection_uid = client.connection.get_uid(connection_details)

        """
        Connections._validate_type(connection_details, u'connection_details', object, True)

        return WMLResource._get_required_element_from_dict(connection_details, u'connection_details',
                                                           [u'metadata', u'id'])

    def list_datasource_types(self):
        """
        List stored datasource types assets.


        **Output**

        .. important::
            This method only prints the list of datasources type in a table format.\n
            **return type**: None\n

        **Example**

         >>> client.connections.list_datasource_types()
        """

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if not self._client.ICP_30 and not self._client.ICP and not self._client.WSD:
            response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                    headers=header_param)
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                    headers=header_param, verify=False)

        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'asset_id'], m[u'entity'][u'type'], m['entity']['status']) for
            m in datasource_details]

        self._list(space_values, [u'NAME', u'DATASOURCE_ID', u'TYPE', u'STATUS'], None, None)

    def list(self):
        """
           List all stored connections.

           **Output**

           .. important::
                This method only prints the list of all connections in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.connections.list()

        """
        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if not self._client.ICP_30 and not self._client.ICP and not self._client.WSD:
            response = requests.get(self._client.service_instance._href_definitions.get_connections_href(),
                                    params=self._client._params(),
                                    headers=header_param)
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_connections_href(),
                                    params=self._client._params(),
                                    headers=header_param, verify=False)

        self._handle_response(200, u'list datasource type', response)

        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'asset_id'],m['metadata']['create_time'], m[u'entity'][u'datasource_type']) for
            m in datasource_details]

        self._list(space_values, [u'NAME', u'ID', u'CREATED',  u'DATASOURCE_TYPE_ID', ], None, None)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_datasource_type_uid_by_name(self, name):
        """
           Get stored datasource types id for the given datasource type name.

           **Parameters**

           .. important::
                #. **name**:  name of datasource type\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the id of given datasource type name.\n
                **return type**: Str\n

           **Example**

            >>> client.connections.get_datasource_type_uid_by_name('cloudobjectstorage')

        """
        Connections._validate_type(name, u'name', str, True)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if not self._client.ICP_30 and not self._client.ICP and not self._client.WSD:
            response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                    headers=header_param)
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                    headers=header_param, verify=False)
        datasource_id = 'None'
        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        for i, ds_resource in enumerate(datasource_details):
            if ds_resource['entity']['name'] == name:
                datasource_id = ds_resource['metadata']['asset_id']
        return datasource_id




import requests
from utils.openshift.cli import get_openshift_token
import urllib.parse


class PrometheusAPI:
    def __init__(self, cluster_username, cluster_password, cluster_server_url):
        # self.username = cluster_username
        # self.password = cluster_password
        self.server_url = cluster_server_url

        token = get_openshift_token(
            username=cluster_username,
            password=cluster_password,
            server=cluster_server_url
        )

        self.session = requests.Session()
        self.session.headers.update({"Authorization": "Bearer {}".format(token)})
        self.session.verify = False

        self.prometheus_url = self.__get_prometheus_url()

    def __get_prometheus_url(self):
        response = self.session.get(
            url="{}/apis/monitoring.coreos.com/v1/prometheuses".format(self.server_url)
        )
        if response.status_code != 200:
            print("Unable to get prometheus url!")
            return None

        prometeus_url = response.json()['items'][0]['spec']['externalUrl']

        print("Prometheus url: {}".format(prometeus_url))
        return prometeus_url

    def get_user_details(self):
        response = self.session.get(
            url="{}/apis/user.openshift.io/v1/users/~".format(self.server_url)
        )
        print(response.status_code)
        print(response.text)

    def execute_query(self, query):
        x_url = self.prometheus_url + "api/v1/query?query=" + query
        response = self.session.get(
            url=x_url
        )
        print(response.text)
        print(response.status_code)

        return response.json()

    def get_cpu_sum_auto_ai(self, auto_ai_experiment_id):
        query_str = "pod:container_cpu_usage:sum{pod=~\"train-wml-autoai-" + str(auto_ai_experiment_id) + ".+\"}[2d]"
        query = urllib.parse.quote(query_str)
        return self.execute_query(query)

    def get_mem_sum_auto_ai(self, auto_ai_experiment_id):
        query_str = "pod:container_memory_usage_bytes:sum{pod=~\"train-wml-autoai-" + str(auto_ai_experiment_id) + ".+\"}[2d]"
        query = urllib.parse.quote(query_str)
        return self.execute_query(query)


# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact import LibrariesArtifact,  LibrariesArtifactReader, LibrariesLoader

logger = logging.getLogger('WmlLibrariesAdapter')


class WmlLibrariesAdapter(object):
    """
    Adapter creating libraries artifact using output from service.
    """

    def __init__(self, library_output, client):

        self.libraries_output = library_output
        self.client = client
        self.libraries_entity = library_output.entity
        self.libraries_metadata = library_output.metadata
        self.uid = library_output.metadata.guid
        self.name = None

    def artifact(self):
        libraries_artifact_builder = type(
            "LibrariesArtifact",
            (LibrariesLoader, LibrariesArtifact, LibrariesArtifactReader, object),
            {}
        )

        prop_map = {
            MetaNames.CREATION_TIME: self.libraries_metadata.created_at,
            MetaNames.LAST_UPDATED: self.libraries_metadata.modified_at,
            MetaNames.LIBRARIES.URL: self.libraries_metadata.url
        }

        if self.libraries_entity.get('name') is not None:
            prop_map[MetaNames.LIBRARIES.NAME] = self.libraries_entity.get('name')
            self.name = self.libraries_entity.get('name')

        if self.libraries_entity.get('description', None) is not None:
            prop_map[MetaNames.LIBRARIES.DESCRIPTION] = self.libraries_entity.get('description',None)

        if self.libraries_entity.get('version') is not None:
            prop_map[MetaNames.LIBRARIES.VERSION] = self.libraries_entity.get('version')

        platform_value = None

        if self.libraries_entity.get('platform') is not None:
            platform_value = self.libraries_entity.get('platform')

        prop_map[MetaNames.LIBRARIES.PLATFORM] = platform_value

        libraries_url = self.libraries_metadata.url
        prop_map[MetaNames.LIBRARIES.URL] = libraries_url
        libraries_content_url = libraries_url + "/content"
        prop_map[MetaNames.LIBRARIES.CONTENT_URL] = libraries_content_url

        local_meta = MetaProps(prop_map)

        libraries_artifact = libraries_artifact_builder(
            uid=self.uid,
            name=self.name,
            meta_props=local_meta
        )

        libraries_artifact.client = self.client
        libraries_artifact._content_href = libraries_content_url

        return libraries_artifact



import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d, create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestAutoAISync,\
    AbstractTestBatch

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, RegressionAlgorithms


class TestAutoAIRemote(AbstractTestAutoAISync, AbstractTestBatch, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """

    cos_resource = None
    data_location = './autoai/data/housing_train.csv'

    data_cos_path = 'data/housing_train.csv'
    batch_cos_filename = "batch_payload_housing.csv"

    SPACE_ONLY = True

    OPTIMIZER_NAME = "housing_train test sdk"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.REGRESSION,
        prediction_column='SalePrice',
        scoring=Metrics.MEAN_ABSOLUTE_ERROR,
        holdout_size=0.18,
        include_only_estimators=[RegressionAlgorithms.SnapBM,
                                 RegressionAlgorithms.SnapRF,
                                 RegressionAlgorithms.SnapDT],
        max_number_of_estimators=3,
    )

    def test_00b_prepare_connection_to_COS(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.data_location,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_00d_prepare_connected_data_asset(self):
        asset_details = self.wml_client.data_assets.store({
            self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.wml_client.data_assets.ConfigurationMetaNames.NAME: "Housing - training asset",
            self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: self.data_cos_path
        })

        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.results_connection = DataConnection(
            location=ContainerLocation()
        )

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_11b_check_snap(self):
        pipeline_params = self.remote_auto_pipelines.get_pipeline_details()
        print(pipeline_params)

        pipeline_nodes = pipeline_params.get('pipeline_nodes')
        self.assertIn('Snap', str(pipeline_nodes), msg=f"{pipeline_nodes}")

    def test_99_delete_connection_and_connected_data_asset(self):
        self.wml_client.data_assets.delete(self.asset_id)
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.data_assets.get_details(self.asset_id)
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()



import unittest,time

import logging
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *

class TestRemoteTrainingSystems(unittest.TestCase):
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        # self.space_id = get_project_id()
        # self.client.set.default_project(self.space_id)

        self.cos_credentials = get_cos_credentials()
        self.cos_resource_crn = self.cos_credentials['resource_instance_id']
        self.instance_crn = get_instance_crn()
        self.space_name = str(uuid.uuid4())

        metadata = {
                     self.client.spaces.ConfigurationMetaNames.NAME: 'client_space_' + self.space_name,
                     self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description',
                     self.client.spaces.ConfigurationMetaNames.STORAGE: {
                         "type": "bmcos_object_storage",
                         "resource_crn": self.cos_resource_crn
                     }
        }

        self.space = self.client.spaces.store(meta_props=metadata, background_mode=False)
        self.space = self.client.spaces.store(meta_props=metadata,  background_mode=False)

        TestRemoteTrainingSystems.space_id = self.client.spaces.get_id(self.space)
        print("space_id: ", TestRemoteTrainingSystems.space_id)
        self.client.set.default_space(TestRemoteTrainingSystems.space_id)

    def test_01_create_remote_training_system(self):

        self.client.remote_training_systems.ConfigurationMetaNames.show()

        meta_props = {
            self.client.remote_training_systems.ConfigurationMetaNames.NAME: "Remote Training Definition",
            self.client.remote_training_systems.ConfigurationMetaNames.TAGS: ["tag1", "tag2"],
            self.client.remote_training_systems.ConfigurationMetaNames.ORGANIZATION: {"name": "name", "region": "EU"},
            self.client.remote_training_systems.ConfigurationMetaNames.ALLOWED_IDENTITIES: [{"id": "43689024", "type": "user"}],
            self.client.remote_training_systems.ConfigurationMetaNames.REMOTE_ADMIN: {"id": "43689024", "type": "user"}
        }

        print(TestRemoteTrainingSystems.space_id)

        rts_details = self.client.remote_training_systems.store(meta_props = meta_props)

        print(rts_details)

        TestRemoteTrainingSystems.rts_asset_id = self.client.remote_training_systems.get_id(rts_details)
        self.assertIsNotNone(TestRemoteTrainingSystems.rts_asset_id)

    def test_02_get_details(self):

        details = self.client.remote_training_systems.get_details(TestRemoteTrainingSystems.rts_asset_id)
        print(details)
        self.assertTrue(TestRemoteTrainingSystems.rts_asset_id in str(details))

    def test_03_list(self):
        self.client.remote_training_systems.list()

    def test_04_revisions(self):
        meta_prop_rts_1_revision1 = {
            self.client.remote_training_systems.ConfigurationMetaNames.NAME: "remote_training_system_update_rev1",
            self.client.remote_training_systems.ConfigurationMetaNames.ORGANIZATION: {'name': 'updated_org_name'}
        }

        rts_update = self.client.remote_training_systems.update(TestRemoteTrainingSystems.rts_asset_id,
                                                                meta_prop_rts_1_revision1)

        print(rts_update)

        self.assertTrue('remote_training_system_update_rev1' in rts_update[u'metadata'][u'name'])
        self.assertTrue('updated_org_name' in rts_update[u'entity'][u'organization'][u'name'])

        # Create revision
        revision = self.client.remote_training_systems.create_revision(TestRemoteTrainingSystems.rts_asset_id)

        print('revision: ', revision)

        self.assertTrue(revision[u'metadata'][u'rev'] == '1')

        # List revisions
        self.client.remote_training_systems.list_revisions(TestRemoteTrainingSystems.rts_asset_id)

    def test_05_delete_rts_asset(self):
        self.client.remote_training_systems.delete(TestRemoteTrainingSystems.rts_asset_id)

    def test_06_delete_space(self):
        self.client.spaces.delete(TestRemoteTrainingSystems.space_id)

if __name__ == '__main__':
    unittest.main()



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from pandas import DataFrame
    from ibm_watson_machine_learning.utils.autoai.enums import PipelineTypes
    from sklearn.pipeline import Pipeline
    from numpy import ndarray

__all__ = [
    "BaseAutoPipelines"
]


class BaseAutoPipelines:
    """
    Base abstract class for Pipeline Optimizers.
    """

    @abstractmethod
    def get_params(self) -> dict:
        """Get configuration parameters of AutoPipelines"""
        pass

    @abstractmethod
    def fit(self, *args, **kwargs) -> 'Pipeline':
        """Run fit job."""
        pass

    @abstractmethod
    def summary(self) -> 'DataFrame':
        """List all computed pipelines."""
        pass

    @abstractmethod
    def get_pipeline_details(self, pipeline_name: str = None) -> dict:
        """Get details of computed pipeline. Details like pipeline steps."""
        pass

    @abstractmethod
    def get_pipeline(self, pipeline_name: str, astype: 'PipelineTypes') -> Union['Pipeline', 'TrainablePipeline']:
        """Get particular computed Pipeline"""
        pass

    @abstractmethod
    def predict(self, X: Union['DataFrame', 'ndarray']) -> 'ndarray':
        """Use predict on top of the computed pipeline."""
        pass



import unittest

try:
    from sklearn.externals import joblib
except ImportError:
    import joblib

from ibm_watson_machine_learning.helpers import pipeline_to_script


class TestPipelineToScript(unittest.TestCase):
    model_path = './autoai/artifacts/bank/bank_model_333.pickle'
    pipeline = None

    @classmethod
    def setUp(cls) -> None:
        cls.pipeline = joblib.load(cls.model_path)

    def test_01__create_python_script(self):
        pipeline_to_script(pipeline=self.pipeline)

        script = None
        try:
            with open("pipeline_script.py") as f:
                script = f.read()

        except FileNotFoundError:
            self.assertIsNone(script, msg="pipeline_script.py file not found!")

        print(script)


if __name__ == '__main__':
    unittest.main()



import unittest
from functools import wraps

from ibm_watson_machine_learning.helpers import DataConnection, S3Connection, S3Location
from ibm_watson_machine_learning.tests.utils import get_cos_credentials, get_wml_credentials, print_test_separators
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.experiment.autoai.optimizers.local_auto_pipelines import LocalAutoPipelines
from lale.operators import TrainedPipeline
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.utils.autoai.enums import PipelineTypes
from sklearn.pipeline import Pipeline
import pandas as pd

"""
Format of COS credentials:

cos_credentials = {
                  "auth_endpoint": "https://iam.cloud.ibm.com/identity/token",
                  "endpoint_url": "https://s3.us.cloud-object-storage.appdomain.cloud",
                  "apikey": "...", 
                  "cos_hmac_keys": {
                    "access_key_id": "...",
                    "secret_access_key": "..."
                  }
                  }
                  
endpoint_url should be appropriate per bucket location, see on UI.
"""


class TestLocalOptimizerWithCOS(unittest.TestCase):
    cos_credentials = None
    bucket = 'nowyobm'
    path = 'bank.csv'
    pipeline_model_hmac = None
    pipeline_model_auth_endpoint = None
    metadata_hmac = None
    metadata_auth_endpoint = None
    pipeline_name = 'Pipeline_1'
    training_status = 'results_wml_autoai/07652672-f643-4b49-b0eb-449fde9d1a20/training-status.json'
    model_location = 'results_wml_autoai/07652672-f643-4b49-b0eb-449fde9d1a20/data/kb/global_output/Pipeline3/model.pickle'

    local_optimizer_hmac: 'LocalAutoPipelines' = None
    local_optimizer: 'LocalAutoPipelines' = None

    data_location = './autoai/data/bank.csv'
    data = None
    X = None
    y = None

    @classmethod
    def setUp(cls) -> None:
        cls.wml_credentials = get_wml_credentials()
        cls.cos_credentials = get_cos_credentials()
        cls.cos_credentials['endpoint_url'] = "https://s3.us-south.cloud-object-storage.appdomain.cloud"
        cls.cos_credentials['auth_endpoint'] = "https://iam.bluemix.net/oidc/token/"
        cls.cos_credentials['api_key'] = '***'
        cls.data = pd.read_csv(cls.data_location)
        cls.X = cls.data.drop(['y'], axis=1)
        cls.y = cls.data['y']

    @print_test_separators
    def test_01a_get_optimizer_with_COS_hmac_credentials(self):
        print("HMAC: Initializing DataConnections with COS credentials and paths...")

        training_data_reference = [DataConnection(
            connection=S3Connection(
                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
            )
        )]

        training_result_reference = DataConnection(
            connection=S3Connection(
                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
                model_location=self.model_location,
                training_status=self.training_status
            )
        )

        TestLocalOptimizerWithCOS.metadata_hmac = dict(
            training_data_reference=training_data_reference,
            training_result_reference=training_result_reference,
            prediction_type='classification',
            prediction_column='y',
            test_size=0.2,
            scoring='roc_auc',
            max_number_of_estimators=1,
        )

        print("HMAC: Initializing AutoAI local scenario with metadata...")

        TestLocalOptimizerWithCOS.local_optimizer_hmac = AutoAI().runs.get_optimizer(
            metadata=self.metadata_hmac
        )

    @print_test_separators
    def test_01b_get_optimizer_with_auth_endpoint_credentials(self):
        print("AUTH: Initializing DataConnections with COS credentials and paths...")

        training_data_reference = [DataConnection(
            connection=S3Connection(
                api_key=self.cos_credentials['apikey'],
                auth_endpoint=self.cos_credentials['auth_endpoint'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
            )
        )]

        training_result_reference = DataConnection(
            connection=S3Connection(
                api_key=self.cos_credentials['apikey'],
                auth_endpoint=self.cos_credentials['auth_endpoint'],
                endpoint_url=self.cos_credentials['endpoint_url']
            ),
            location=S3Location(
                bucket=self.bucket,
                path=self.path,
                model_location=self.model_location,
                training_status=self.training_status
            )
        )

        TestLocalOptimizerWithCOS.metadata_auth_endpoint = dict(
            training_data_reference=training_data_reference,
            training_result_reference=training_result_reference,
            prediction_type='classification',
            prediction_column='y',
            test_size=0.2,
            scoring='roc_auc',
            max_number_of_estimators=1,
        )

        print("AUTH: Initializing AutoAI local scenario with metadata...")
        TestLocalOptimizerWithCOS.local_optimizer = AutoAI().runs.get_optimizer(
            metadata=self.metadata_auth_endpoint
        )

    @print_test_separators
    def test_02a_get_pipeline_hmac(self):
        print("HMAC: Fetching and store pipeline by name, LALE...")

        pipeline_lale = self.local_optimizer_hmac.get_pipeline(
            pipeline_name=self.pipeline_name, persist=True)
        self.assertIsInstance(pipeline_lale, TrainedPipeline, msg="Loaded model should be of type TrainedPipeline")

        print("HMAC: Fetching pipeline by name, SKLEARN...")

        pipeline_skl = self.local_optimizer_hmac.get_pipeline(
            pipeline_name=self.pipeline_name, astype=PipelineTypes.SKLEARN)
        self.assertIsInstance(pipeline_skl, Pipeline, msg="Loaded model should be of type SKLEARN")

        print("HMAC: Fetching best pipeline, LALE...")

        try:
            TestLocalOptimizerWithCOS.pipeline_model_hmac = self.local_optimizer_hmac.get_pipeline()
        except Exception as e:
            self.assertIsInstance(self.pipeline_model_hmac, TrainedPipeline,
                                  msg=f"Cannot load best pipeline model, Error: {e}")

    @print_test_separators
    def test_02b_get_pipeline_auth_endpoint(self):
        print("AUTH: Fetching and store pipeline by name, LALE...")

        pipeline_lale = self.local_optimizer.get_pipeline(pipeline_name=self.pipeline_name, persist=True)
        self.assertIsInstance(pipeline_lale, TrainedPipeline, msg="Loaded model should be of type TrainedPipeline")

        print("AUTH: Fetching pipeline by name, SKLEARN...")

        pipeline_skl = self.local_optimizer.get_pipeline(pipeline_name=self.pipeline_name, astype=PipelineTypes.SKLEARN)
        self.assertIsInstance(pipeline_skl, Pipeline, msg="Loaded model should be of type SKLEARN")

        print("AUTH: Fetching best pipeline, LALE...")

        try:
            TestLocalOptimizerWithCOS.pipeline_model_auth_endpoint = self.local_optimizer.get_pipeline()
        except Exception as e:
            self.assertIsInstance(self.pipeline_model_auth_endpoint, TrainedPipeline,
                                  msg=f"Cannot load best pipeline model, Error: {e}")

    @print_test_separators
    def test_03a_get_params__fetch_dict_with_parameters_hmac(self):
        print("HMAC: Getting optimizer parameters...")

        params = self.local_optimizer_hmac.get_params()
        print(params)

    @print_test_separators
    def test_03b_get_params__fetch_dict_with_parameters_auth_endpoint(self):
        print("AUTH: Getting optimizer parameters...")

        params = self.local_optimizer.get_params()
        print(params)

    @print_test_separators
    def test_04a_fit__not_implemented_hmac(self):
        print("HMAC: Calling fit() method...")

        with self.assertRaises(NotImplementedError, msg="\"fit\" should raise NotImplementedError"):
            self.local_optimizer_hmac.fit(X=self.X, y=self.y)

    @print_test_separators
    def test_04b_fit__not_implemented_auth_endpoint(self):
        print("AUTH: Calling fit() method...")

        with self.assertRaises(NotImplementedError, msg="\"fit\" should raise NotImplementedError"):
            self.local_optimizer.fit(X=self.X, y=self.y)

    @print_test_separators
    def test_05a_summary__return_data_frame_with_summary_hmac(self):
        print("HMAC: Fetching summary of experiment...")

        summary = self.local_optimizer_hmac.summary()
        print(summary)

    @print_test_separators
    def test_05b_summary__return_data_frame_with_summary_auth_endpoint(self):
        print("AUTH: Fetching summary of experiment...")
        summary = self.local_optimizer.summary()
        print(summary)

    @print_test_separators
    def test_06a_get_pipeline_details_hmac(self):
        print("HMAC: Fetching best pipeline details...")

        best_pipeline_details = self.local_optimizer_hmac.get_pipeline_details()
        print(f"best pipeline details: {best_pipeline_details}")

        print("HMAC: Fetching pipeline details...")

        pipeline_details = self.local_optimizer_hmac.get_pipeline_details(pipeline_name=self.pipeline_name)
        print(f"pipeline details: {pipeline_details}")

    @print_test_separators
    def test_06b_get_pipeline_details_auth_endpoint(self):
        print("AUTH: Fetching best pipeline details...")

        best_pipeline_details = self.local_optimizer.get_pipeline_details()
        print(f"best pipeline details: {best_pipeline_details}")

        print("AUTH: Fetching pipeline details...")

        pipeline_details = self.local_optimizer.get_pipeline_details(pipeline_name=self.pipeline_name)
        print(f"pipeline details: {pipeline_details}")

    @print_test_separators
    def test_07a_predict__on_best_pipeline_hmac(self):
        print("HMAC: Calling predict on the best pipeline...")

        predictions = self.local_optimizer_hmac.predict(X=self.X)
        print(predictions)

    @print_test_separators
    def test_07b_predict__on_best_pipeline_auth_endpoint(self):
        print("AUTH: Calling predict on the best pipeline...")

        predictions = self.local_optimizer.predict(X=self.X)
        print(predictions)

    @print_test_separators
    def test_08a_get_data_connections__return_training_data_connection_and_recreate_holdout_split_hmac(self):
        print("HMAC: Reading training data with holdout split...")

        training_df, holdout_df = self.local_optimizer_hmac.get_data_connections()[0].read(with_holdout_split=True)
        print(f"Training data frame: {training_df}")
        print(f"Holdout data frame: {holdout_df}")

    @print_test_separators
    def test_08b_get_data_connections__return_training_data_connection_and_recreate_holdout_split_auth_endpoint(self):
        print("AUTH: Reading training data with holdout split...")

        training_df, holdout_df = self.local_optimizer.get_data_connections()[0].read(with_holdout_split=True)
        print(f"Training data frame: {training_df}")
        print(f"Holdout data frame: {holdout_df}")

    # @print_test_separators
    # def test_09a_deploy_model_from_object_hmac(self):
    #     print("HMAC: Deploying model from object...")
    #
    #     # service = WebService(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
    #     service = Batch(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
    #     service.create(
    #         model=self.pipeline_model_hmac,
    #         metadata=self.metadata_hmac,
    #         deployment_name="Test deployment from auto-gen notebook and COS"
    #     )
    #
    #     print(service)
    #     print(service.get_params())
    #     predictions = service.score(payload=self.X.iloc[:10])
    #     print(predictions)
    #     print(service.list())
    #     print(service.delete())
    #     print(service.list())
    #
    # @print_test_separators
    # def test_09b_deploy_model_from_object_auth_endpoint(self):
    #     print("AUTH: Deploying model from object...")
    #
    #     # service = WebService(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
    #     service = Batch(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
    #     service.create(
    #         model=self.pipeline_model_auth_endpoint,
    #         metadata=self.metadata_auth_endpoint,
    #         deployment_name="Test deployment from auto-gen notebook and COS"
    #     )
    #
    #     print(service)
    #     print(service.get_params())
    #     predictions = service.score(payload=self.X.iloc[:10])
    #     print(predictions)
    #     print(service.list())
    #     print(service.delete())
    #     print(service.list())

    @print_test_separators
    def test_10a_deploy_model_from_pipeline_name_hmac(self):
        print("HMAC: Deploying model from pipeline name...")

        # service = WebService(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
        service = Batch(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
        service.create(
            model=self.pipeline_name,
            metadata=self.metadata_hmac,
            deployment_name="Test deployment from auto-gen notebook and COS"
        )

        print(service)
        print(service.get_params())
        predictions = service.score(payload=self.X.iloc[:10])
        print(predictions)
        print(service.list())
        print(service.delete())
        print(service.list())

    @print_test_separators
    def test_10b_deploy_model_from_pipeline_name_auth_endpoint(self):
        print("AUTH: Deploying model from pipeline name...")

        # service = WebService(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
        service = Batch(self.wml_credentials, space_id='4e0541b8-174b-4d04-ae21-14c15475d68a')
        service.create(
            model=self.pipeline_name,
            metadata=self.metadata_auth_endpoint,
            deployment_name="Test deployment from auto-gen notebook and COS"
        )

        print(service)
        print(service.get_params())
        predictions = service.score(payload=self.X.iloc[:10])
        print(predictions)
        print(service.list())
        print(service.delete())
        print(service.list())


if __name__ == '__main__':
    unittest.main()



import unittest
from sklearn.pipeline import Pipeline
from pprint import pprint
import pandas as pd
import traceback
from os import environ

from ibm_watson_machine_learning import APIClient

from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.workspace import WorkSpace

from ibm_watson_machine_learning.helpers.connections import S3Connection, S3Location, DataConnection, FSLocation

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines

from ibm_watson_machine_learning.utils.autoai.errors import NotExistingCOSResource

from ibm_watson_machine_learning.tests.utils import (get_wml_credentials, get_cos_credentials, bucket_exists,
                                                     create_bucket, get_space_id, create_connection_to_cos)
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment

from ibm_watson_machine_learning.utils.autoai.enums import PositiveLabelClass, TShirtSize, Transformers


class TestAutoAIRemote(unittest.TestCase):
    """
    The test can be run on CLOUD, WMLS and CPD (not tested)
    The test covers:
    - COS set-up (if run on Cloud): checking if bucket exists for the cos instance, if not new bucket is create
    - Saving data `/bank.cdv` to COS/data assets
    - downloading training data from cos/data assets
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    """
    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    cos_credentials = None
    pipeline_opt: 'RemoteAutoPipelines' = None
    service: 'WebService' = None
    service_batch: 'Batch' = None

    data_location = './autoai/data/iris_dataset.csv'

    prediction_column = 'species'

    trained_pipeline_details = None
    run_id = None

    data_connection = None
    results_connection = None

    train_data = None

    bucket_name = environ.get('BUCKET_NAME', "wml-autoaitests-qa")
    pod_version = environ.get('KB_VERSION', None)
    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')

    cos_endpoint = "https://s3.us-south.cloud-object-storage.appdomain.cloud"
    data_cos_path = 'data/iris_dataset.csv'
    cos_resource_instance_id = None

    results_cos_path = 'results_wml_autoai'

    pipeline: 'Pipeline' = None
    lale_pipeline = None
    deployed_pipeline = None
    hyperopt_pipelines = None
    new_pipeline = None
    new_sklearn_pipeline = None
    X_values = None
    y_values = None

    OPTIMIZER_NAME = 'IRIS csv wml_autoai multiclass test'
    DEPLOYMENT_NAME = "IRIS AutoAI Deployment tests"

    project_id = None
    space_id = None

    asset_id = None

    row_no_with_drop = None
    row_no_without_drop = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        if not cls.wml_client.ICP:
            cls.cos_credentials = get_cos_credentials()
            cls.cos_endpoint = cls.cos_credentials.get('endpoint_url')
            cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        cls.project_id = cls.wml_credentials.get('project_id')

    def test_00a_space_cleanup(self):
        space_cleanup(self.wml_client,
                      get_space_id(self.wml_client, self.space_name,
                                   cos_resource_instance_id=self.cos_resource_instance_id),
                      days_old=7)
        TestAutoAIRemote.space_id = get_space_id(self.wml_client, self.space_name,
                                                 cos_resource_instance_id=self.cos_resource_instance_id)
        if self.wml_client.ICP:
            self.wml_client.set.default_project(self.project_id)
        else:
            self.wml_client.set.default_space(self.space_id)

    def test_00b_prepare_COS_instance(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.data_location,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

        print(f"Using COS bucket: {TestAutoAIRemote.bucket_name}")

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        TestAutoAIRemote.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                             project_id=self.project_id,
                                             space_id=self.space_id)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_save_remote_data_and_DataConnection_setup(self):
        if self.wml_client.ICP:
            TestAutoAIRemote.data_connection = DataConnection(
                location=FSLocation(path=self.data_location))
            TestAutoAIRemote.results_connection = None

        else:  # for cloud and COS
            TestAutoAIRemote.data_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.data_cos_path)
            )
            TestAutoAIRemote.results_connection = DataConnection(
                connection=S3Connection(endpoint_url=self.cos_endpoint,
                                        access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                                        secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']),
                location=S3Location(bucket=self.bucket_name,
                                    path=self.results_cos_path)
            )
            TestAutoAIRemote.data_connection.write(data=self.data_location, remote_name=self.data_cos_path)

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)

    def test_02a_read_saved_remote_data_before_fit(self):
        data = self.data_connection.read()
        print(f"Data sample")
        print(data.head())
        self.assertGreater(len(data), 0)

    def test_03_initialize_optimizer_with_drop_duplicates(self):
        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            desc='test description',
            prediction_type=self.experiment.PredictionType.MULTICLASS,
            prediction_column=self.prediction_column,
            scoring=PositiveLabelClass.PRECISION_SCORE_MICRO,
            cognito_transform_names=[Transformers.DIFF, Transformers.SUM, Transformers.MAX,
                                     Transformers.SIN, Transformers.SUM],
            holdout_size=0.1,
            max_number_of_estimators=1,
            t_shirt_size=TShirtSize.S if self.wml_client.ICP else TShirtSize.L,
            autoai_pod_version=self.pod_version
        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_04_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)
        assert ('drop_duplicates' in parameters)
        assert (parameters['drop_duplicates'])
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_05_fit_run_training_of_auto_ai_in_wml(self):
        TestAutoAIRemote.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=[self.data_connection],
            training_results_reference=self.results_connection,
            background_mode=False)

        TestAutoAIRemote.run_id = self.trained_pipeline_details['metadata']['id']

        self.assertIsNotNone(self.data_connection.auto_pipeline_params,
                             msg='DataConnection auto_pipeline_params was not updated.')

        step = self.trained_pipeline_details['entity']['status']['metrics'][0]['context']['intermediate_model'][
            'composition_steps'][0]
        TestAutoAIRemote.row_no_with_drop = int(step.split('_')[2])
        print(TestAutoAIRemote.row_no_with_drop)
        self.assertEqual(TestAutoAIRemote.row_no_with_drop, 147)

    def test_06_initialize_optimizer_without_drop_duplicates(self):
        TestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            desc='test description',
            prediction_type=self.experiment.PredictionType.MULTICLASS,
            prediction_column=self.prediction_column,
            scoring=PositiveLabelClass.PRECISION_SCORE_MICRO,
            cognito_transform_names=[Transformers.DIFF, Transformers.SUM, Transformers.MAX,
                                     Transformers.SIN, Transformers.SUM],
            holdout_size=0.1,
            max_number_of_estimators=1,
            t_shirt_size=TShirtSize.S if self.wml_client.ICP else TShirtSize.L,
            drop_duplicates=False
        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_07_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)
        assert ('drop_duplicates' in parameters)
        assert (not parameters['drop_duplicates'])
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_08_fit_run_training_of_auto_ai_in_wml(self):
        TestAutoAIRemote.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=[self.data_connection],
            training_results_reference=self.results_connection,
            background_mode=False)

        TestAutoAIRemote.run_id = self.trained_pipeline_details['metadata']['id']

        self.assertIsNotNone(self.data_connection.auto_pipeline_params,
                             msg='DataConnection auto_pipeline_params was not updated.')

        step = self.trained_pipeline_details['entity']['status']['metrics'][0]['context']['intermediate_model'][
            'composition_steps'][0]
        TestAutoAIRemote.row_no_without_drop = int(step.split('_')[2])
        print(TestAutoAIRemote.row_no_without_drop)
        self.assertEqual(TestAutoAIRemote.row_no_without_drop, 150)

    def test_09_compare(self):
        print(TestAutoAIRemote.row_no_with_drop, TestAutoAIRemote.row_no_without_drop)
        self.assertNotEqual(TestAutoAIRemote.row_no_with_drop, TestAutoAIRemote.row_no_without_drop)


if __name__ == '__main__':
    unittest.main()
